@@ -1,41 +1,33 @@
 from __future__ import annotations
-from dataclasses import dataclass
+
 import re
 import os
 import json
-from pathlib import Path
 import glob
+
+from pathlib import Path
 from functools import cached_property
-from io import TextIOWrapper
-from typing import Union
-from send2trash import send2trash
-import copy
+from typing import Union, Tuple
+
 import dpath.util
 
 NO_ARGUMENT = object()
 TEXTURE_EXTENSIONS = ["png", "jpg", "jpeg", "tga"]
 SOUND_EXTENSIONS = ["wav", "fsb", "ogg"]
 
-
 def create_nested_directory(path: str):
     """
     Creates a nested directory structure if it doesn't exist.
     """
     Path(path).parent.mkdir(parents=True, exist_ok=True)
 
-
-def freeze(o):
+def save_json(file_path, data):
     """
-    Makes a hash out of anything that contains only list,dict and hashable types
-    including string and numeric types
+    Saves json to a file_path, creating nested directory if required.
     """
-    if isinstance(o, dict):
-        return frozenset({ k:freeze(v) for k,v in o.items()}.items())
-
-    if isinstance(o,list):
-        return tuple(freeze(v) for v in o)
-
-    return hash(o)
+    create_nested_directory(file_path)
+    with open(file_path, "w+") as file_head:
+        return json.dump(data, file_head, indent=2, ensure_ascii=False)
 
 
 def convert_to_notify_structure(data: Union[dict, list], parent: Resource) -> Union[NotifyDict, NotifyList]:
@@ -214,23 +206,6 @@ class Resource():
         """
         self._resources.append(resource)
 
-    def _should_save(self):
-        """
-        Whether the asset can be saved.
-        By default, this is related to dirty status
-
-        Can be overridden by subclasses.
-        """
-        return self.dirty
-
-    def _should_delete(self):
-        """
-        Whether the asset can be deleted.
-
-        By default, returns true.
-        """
-        return True
-
     def _save(self):
         """
         Internal implementation of asset saving.
@@ -256,12 +231,12 @@ class Resource():
         the internal _save implementation.
         """
 
-        # TODO Maybe only files need this check?
-        # Assets without packs may not save.
-        if not self.pack:
+        # File Assets without packs may not save.
+        if isinstance(self, FileResource) and not self.pack:
             raise FloatingAssetError("Assets without a pack cannot be saved.")
 
-        if self._should_save() or force:
+        # Only dirty assets can be saved, unless forced.
+        if self.dirty or force:
             self.dirty = False
 
             # Save all children first
@@ -271,22 +246,17 @@ class Resource():
             # Now, save this resource
             self._save()
 
-    def delete(self, force=False):
+    def delete(self):
         """
         Deletes the resource. Should not be overridden.
         """
 
-        # TODO: Does deletion really need to be recursive?
-        if self._should_delete() or force:
-            # First, delete all resources of children
-            for resource in self._resources:
-                resource.delete(force=force)
-
-            # Then delete self
-            self._delete()
-
-            # TODO Do we need to call save here?
+        # First, delete all resources of children
+        for resource in self._resources:
+            resource.delete()
 
+        # Then delete self
+        self._delete()
 
 class FileResource(Resource):
     """
@@ -310,8 +280,6 @@ class FileResource(Resource):
         if file_path:
             self.file_name = os.path.basename(file_path)
 
-        # Protected
-        self._hash = self._create_hash()
         self._mark_for_deletion: bool = False
 
     def _delete(self):
@@ -323,26 +291,6 @@ class FileResource(Resource):
         """
         self._mark_for_deletion = True
 
-    def _should_delete(self):
-        return not self._mark_for_deletion and super()._should_delete()
-
-    def _should_save(self):
-        # Files that have been deleted cannot be saved
-        if self._mark_for_deletion:
-            return False
-
-        return self._hash != self._create_hash() or super()._should_save()
-
-    def _create_hash(self):
-        """
-        Creates a hash for the file.
-
-        If a file doesn't want to implement a hash, then the file will always
-        be treated as the same, allowing other checks to take control.
-        """
-        return 0
-
-
 class JsonResource(Resource):
     """
     Parent class, which is responsible for all resources which contain
@@ -356,7 +304,7 @@ class JsonResource(Resource):
     def __init__(self, data: dict = None, file: FileResource = None, pack: Pack = None) -> None:
         super().__init__(file=file, pack=pack)
         self._data = convert_to_notify_structure(data, self)
-
+    
     @property
     def data(self):
         return self._data
@@ -438,7 +386,6 @@ class JsonResource(Resource):
                 f"Path {json_path} does not exist."
             ) from exception
 
-
     def get_data_at(self, json_path):
         """
         Returns a list of jsonpaths found at this jsonpath location.
@@ -483,19 +430,55 @@ class JsonFileResource(FileResource, JsonResource):
         if data is not None:
             self.data = convert_to_notify_structure(data, self)
         else:
-            self.data = convert_to_notify_structure(self.pack.load_json(self.file_path), self)
+            self.data = convert_to_notify_structure(self.load_json(self.file_path), self)
 
         # Init json resource, which relies on the new data attribute
         JsonResource.__init__(self, data=self.data, file=self, pack=pack)
 
+    def load_json(self, local_path: str) -> dict:
+        """
+        Loads json from file. `local_path` paramater is local to the projects
+        input path.
+        """
+        file_path = os.path.join(self.pack.input_path, local_path)
+        if not os.path.exists(file_path):
+            raise AssetNotFoundError(f"File not found: {file_path}")
+        
+        try:
+            with open(file_path, "r", encoding='utf8') as fh:
+                try:
+                    return json.load(fh)
+                except json.JSONDecodeError:
+                    try:
+                        fh.seek(0)
+                        contents = ""
+                        for line in fh.readlines():
+                            cleaned_line = line.split("//", 1)[0]
+                            if len(cleaned_line) > 0 and line.endswith("\n") and "\n" not in cleaned_line:
+                                cleaned_line += "\n"
+                            contents += cleaned_line
+                        while "/*" in contents:
+                            pre_comment, post_comment = contents.split("/*", 1)
+                            contents = pre_comment + post_comment.split("*/", 1)[1]
+                        return json.loads(contents)
+                    except json.JSONDecodeError as exception:
+                        return {}
+        except Exception:
+            raise InvalidJsonError(file_path)
+
     def _save(self):
-        self.pack.save_json(self.file_path, self.data)
+        save_path = os.path.join(self.pack.output_path, self.file_path)
+        create_nested_directory(save_path)
 
-    def _delete(self):
-        self._mark_for_deletion = True
-
-    def _create_hash(self):
-        return freeze(self.data)
+        # If the file has been marked for deletion, delete it
+        if self._mark_for_deletion:
+            # If the paths are the same, delete the file, otherwise
+            # we can just pass
+            if smart_compare(self.pack.input_path, self.pack.output_path):
+                os.remove(save_path)
+        else:
+            with open(save_path, "w+") as file_head:
+                return json.dump(self.data, file_head, indent=2, ensure_ascii=False)
 
 
 class JsonSubResource(JsonResource):
@@ -629,6 +612,10 @@ class ModelTriple(JsonSubResource):
         return self.parent.pack.get_model(self.identifier)
     
     def exists(self):
+        """
+        Whether the model described by this triple exists.
+        Can be used to detect if a entity is referencing a missing model.
+        """
         try:
             self.parent.pack.get_model(self.identifier)
             return True
@@ -663,6 +650,10 @@ class AnimationTriple(JsonSubResource):
         return self.parent.pack.get_animation(self.identifier)
     
     def exists(self):
+        """
+        Whether the animation described by this triple exists.
+        Can be used to detect if a entity is referencing a missing animation.
+        """
         try:
             self.parent.pack.get_animation(self.identifier)
             return True
@@ -732,16 +723,16 @@ class MaterialTriple(JsonSubResource):
         except AssetNotFoundError:
             return False
 
-@dataclass
 class Translation:
     """
     Dataclass for a translation. Many translations together make up a
     TranslationFile.
     """
-    key: str
-    value: str
-    comment: str
 
+    def __init__(self, key: str, value: str, comment: str = "") -> None:
+        self.key = key
+        self.value = value
+        self.comment = comment
 
 class Command(Resource):
     """
@@ -835,6 +826,15 @@ class LanguageFile(FileResource):
     def __init__(self, file_path: str = None, pack: Pack = None) -> None:
         super().__init__(file_path=file_path, pack=pack)
         self.__translations: list[Translation] = []
+    
+    def get_translation(self, key: str) -> Translation:
+        """
+        Whether the language file contains the specified key.
+        """
+        for translation in self.translations:
+            if translation.key == key:
+                return translation 
+        raise AssetNotFoundError(f"Translation with key '{key}' not found in language file '{self.file_path}'.")
 
     def contains_translation(self, key: str) -> bool:
         """
@@ -912,6 +912,12 @@ class Pack():
         # explicitely saved.
         self.output_path: str = input_path
 
+    def set_save_locations(self, save_location: str) -> None:
+        """
+        Sets the output path for this pack.
+        """
+        self.output_path = save_location
+
     @cached_property
     def project(self) -> Project:
         """
@@ -919,31 +925,6 @@ class Pack():
         """
         return self._project
 
-    def load_json(self, local_path: str) -> dict:
-        """
-        Loads json from file. `local_path` paramater is local to the projects
-        input path.
-        """
-        file_path = os.path.join(self.input_path, local_path)
-        if not os.path.exists(file_path):
-            raise AssetNotFoundError(f"File not found: {file_path}")
-        
-        return self.get_json_from_path(file_path)
-
-    def save_json(self, local_path, data):
-        """
-        Saves json to file. 'local_path' paramater is local to the projects
-        output path.
-        """
-        return self.__save_json(os.path.join(self.output_path, local_path), data)
-
-    def delete_file(self, local_path):
-        # TODO: Why is this local to output path? Why does this except pass?
-        try:
-            send2trash(os.path.join(self.output_path, local_path))
-        except Exception:
-            pass
-
     def save(self, force=False):
         """
         Saves every child resource.
@@ -967,10 +948,10 @@ class Pack():
     def get_language_file(self, file_path:str) -> LanguageFile:
         """
         Gets a specific language file, based on the name of the language file.
-        For example, 'en_GB.lang'
+        For example, 'texts/en_GB.lang'
         """
         for language_file in self.language_files:
-            if language_file.file_path == file_path:
+            if smart_compare(language_file.file_path, file_path):
                 return language_file
         raise AssetNotFoundError(file_path)
 
@@ -986,50 +967,6 @@ class Pack():
             
         return self._language_files
 
-    ## TODO: Move these static methods OUT
-    @staticmethod
-    def get_json_from_path(path:str) -> dict:
-        try:
-            with open(path, "r", encoding='utf8') as fh:
-                return Pack.__get_json_from_file(fh)
-        except Exception:
-            raise InvalidJsonError(path)
-
-    @staticmethod
-    def __get_json_from_file(fh:TextIOWrapper) -> dict:
-        """
-        Loads json from file. First attempts to load with fast json.load method,
-        but if this fails, a custom comment-aware scraper is used.
-        """
-        try:
-            return json.load(fh)
-        except json.JSONDecodeError:
-            try:
-                fh.seek(0)
-                contents = ""
-                for line in fh.readlines():
-                    cleaned_line = line.split("//", 1)[0]
-                    if len(cleaned_line) > 0 and line.endswith("\n") and "\n" not in cleaned_line:
-                        cleaned_line += "\n"
-                    contents += cleaned_line
-                while "/*" in contents:
-                    pre_comment, post_comment = contents.split("/*", 1)
-                    contents = pre_comment + post_comment.split("*/", 1)[1]
-                return json.loads(contents)
-            except json.JSONDecodeError as exception:
-                return {}
-
-    @staticmethod
-    def __save_json(file_path, data):
-        """
-        Saves json to a file_path, creating nested directory if required.
-        """
-        create_nested_directory(file_path)
-
-        with open(file_path, "w+") as file_head:
-            return json.dump(data, file_head, indent=2, ensure_ascii=False)
-
-
 class Project():
     """
     A Project is a holder class which contains reference to both a single
@@ -1040,7 +977,29 @@ class Project():
         self.__resource_path = resource_path
         self.__resource_pack = None
         self.__behavior_pack : BehaviorPack = None
+    
+    def set_save_locations(self, save_location: str) -> None:
+        """
+        Sets the save location of the RP and the BP, based on the folder
+        name from their input path.
 
+        In other words, pass in a folder where you want both the RP and the BP 
+        to be saved.
+
+        If you need finer control, use the 'set_save_location' method on the 
+        ResourcePack and BehaviorPack instead.
+        """
+        self.resource_pack.output_path = save_location + "/" + os.path.dirname(self.resource_pack.input_path)
+        self.behavior_pack.output_path = save_location + "/" + os.path.dirname(self.behavior_pack.input_path)
+
+    def get_packs(self) -> Tuple[behavior_pack, resource_pack]:
+        """
+        Returns Behavior Pack followed by ResourcePack.
+        Useful for quickly defining rp and bp:
+        rp, bp = Project("...").get_packs()
+        """
+        return self.behavior_pack, self.resource_pack
+        
     @cached_property
     def resource_pack(self) -> ResourcePack:
         """
@@ -1081,11 +1040,11 @@ class ResourcePack(Pack):
         self.__textures: list[str] = []
         self.__material_files: list[MaterialFile] = []
         self.__materials: list[Material] = []
-
+        self.__fogs: list[FogFile] = []
         self.__sounds_file: SoundsFile = None
         self.__sound_definitions_file: SoundDefinitionsFile = None
-        self.__terrain_texture_file: TerrainTextureFile = None
-        self.__item_texture_file: ItemTextureFile = None
+        self.__terrain_texture_file: StandAloneTextureFile = None
+        self.__item_texture_file: StandAloneTextureFile = None
         self.__flipbook_textures_file: FlipbookTexturesFile = None
         self.__blocks_file: BlocksFile = None
         self.__biomes_client_file: BiomesClientFile = None
@@ -1141,9 +1100,16 @@ class ResourcePack(Pack):
         for local_path in glob.glob(base_directory + "/**/*.json", recursive=True):
             local_path = os.path.relpath(local_path, self.input_path)
             self.__entities.append(EntityFileRP(file_path = local_path, pack = self))
-            
         return self.__entities
 
+    @cached_property
+    def fogs(self) -> list[FogFile]:
+        base_directory = os.path.join(self.input_path, "fogs")
+        for local_path in glob.glob(base_directory + "/**/*.json", recursive=True):
+            local_path = os.path.relpath(local_path, self.input_path)
+            self.__fogs.append(FogFile(file_path = local_path, pack = self))
+        return self.__fogs
+
     @cached_property
     def model_files(self) -> list[ModelFile]:
         base_directory = os.path.join(self.input_path, "models")
@@ -1287,15 +1253,15 @@ class ResourcePack(Pack):
         return self.__sound_definitions_file
 
     @cached_property
-    def terrain_texture_file(self) -> TerrainTextureFile:
+    def terrain_texture_file(self) -> StandAloneTextureFile:
         file_path = os.path.join("textures", "terrain_texture.json")
-        self.__terrain_texture_file = TerrainTextureFile(file_path = file_path, pack = self)
+        self.__terrain_texture_file = StandAloneTextureFile(file_path = file_path, pack = self)
         return self.__terrain_texture_file
 
     @cached_property
-    def item_texture_file(self) -> ItemTextureFile:
+    def item_texture_file(self) -> StandAloneTextureFile:
         file_path = os.path.join("textures", "item_texture.json")
-        self.__item_texture_file = ItemTextureFile(file_path = file_path, pack = self)
+        self.__item_texture_file = StandAloneTextureFile(file_path = file_path, pack = self)
         return self.__item_texture_file
 
     @cached_property
@@ -1323,6 +1289,13 @@ class ResourcePack(Pack):
                 return child
         raise AssetNotFoundError(identifier)
 
+    def get_fog(self, identifier:str) -> FogFile:
+        for fog in self.fogs:
+            if smart_compare(fog.identifier, identifier):
+                return fog
+        raise AssetNotFoundError(f"Fog with identifier {identifier} could not be found.")
+
+
     def get_attachable(self, identifier:str) -> AttachableFileRP:
         """
         Gets an AttachableFileRP by its identifier.
@@ -1918,23 +1891,32 @@ class BiomesClientFile(JsonFileResource):
     def __init__(self, data: dict = None, file_path: str = None, pack: Pack = None) -> None:
         super().__init__(data = data, file_path = file_path, pack = pack)
 
-
-class ItemTextureFile(JsonFileResource):
+class StandAloneTextureFile(JsonFileResource):
     """
-    ItemTextureFile is a class which represents the data stored in 'rp/textures/item_texture.json'
+    StandAloneTextureFile is a class which represents the data stored in 'rp/textures/*_texture.json'
+    style files.
     """
     def __init__(self, data: dict = None, file_path: str = None, pack: Pack = None) -> None:
         super().__init__(data = data, file_path = file_path, pack = pack)
+        self.__texture_definitions = []
     
+    @cached_property
+    def texture_definitions(self) -> list[TextureFileDouble]:
+        for path, data in self.get_data_at("texture_data"):
+            self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = path, data = data))
+        return self.__texture_definitions
 
+    def get_texture_definition(self, shortname: str) -> TextureFileDouble:
+        for child in self.texture_definitions:
+            if child.shortname == shortname:
+                return child
+        raise AssetNotFoundError(f"Texture definition for shortname '{shortname}' not found.")
 
-class TerrainTextureFile(JsonFileResource):
-    """
-    TerrainTextureFile is a class which represents the data stored in 'rp/textures/terrain_texture.json'
-    """
-    def __init__(self, data: dict = None, file_path: str = None, pack: Pack = None) -> None:
-        super().__init__(data = data, file_path = file_path, pack = pack)
-
+    def add_texture_definition(self, shortname: str, textures: list[str]):
+        self.set_jsonpath(f"texture_data/{shortname}", {
+            "textures": textures
+        })
+        self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = f"texture_data/{shortname}", data = {"textures": textures}))
 
 class SoundsFile(JsonFileResource):
     """
@@ -2229,11 +2211,78 @@ class EntityFileBP(JsonFileResource):
         return new_object
 
 
+class FogFile(JsonFileResource):
+    def __init__(self, data: dict = None, file_path: str = None, pack: Pack = None) -> None:
+        super().__init__(data = data, file_path = file_path, pack = pack)
+        self.__distance_components: list[Component] = []
+        self.__volumetric_density_components: list[Component] = []
+        self.__volumetric_media_coefficients: list[Component] = []
+
+    @property
+    def format_version(self):
+        return self.get_jsonpath("format_version")
+    
+    @format_version.setter
+    def format_version(self, format_version):
+        return self.set_jsonpath("format_version", format_version)
+
+    @property
+    def identifier(self):
+        return self.get_jsonpath("minecraft:fog_settings/description/identifier")
+    
+    @format_version.setter
+    def format_version(self, identifier):
+        return self.set_jsonpath("minecraft:fog_settings/description/identifier", identifier)
+
+    @cached_property
+    def distance_components(self) -> list[Component]:
+        for path, data in self.get_data_at("minecraft:fog_settings/distance"):
+            self.__distance_components.append(Component(parent = self, json_path = path, data = data))
+        return self.__distance_components
+
+    @cached_property
+    def volumetric_density_components(self) -> list[Component]:
+        for path, data in self.get_data_at("minecraft:fog_settings/volumetric/density"):
+            self.__volumetric_density_components.append(Component(parent = self, json_path = path, data = data))
+        return self.__volumetric_density_components
+
+    @cached_property
+    def volumetric_media_coefficients(self) -> list[Component]:
+        for path, data in self.get_data_at("minecraft:fog_settings/volumetric/media_coefficients"):
+            self.__volumetric_media_coefficients.append(Component(parent = self, json_path = path, data = data))
+        return self.__volumetric_media_coefficients
+
+    def get_distance_component(self, id:str) -> Component:
+        for child in self.distance_components:
+            if smart_compare(child.id, id):
+                return child
+        raise AssetNotFoundError(f"Distance component with id '{id}' not found.")
+
+    def get_volumetric_density_component(self, id:str) -> Component:
+        for child in self.distance_components:
+            if smart_compare(child.id, id):
+                return child
+        raise AssetNotFoundError(f"Volumetric density component with id '{id}' not found.")
+
+    def get_volumetric_media_coefficient(self, id:str) -> Component:
+        for child in self.volumetric_media_coefficients:
+            if smart_compare(child.id, id):
+                return child
+        raise AssetNotFoundError(f"volumetric media coefficient component with id '{id}' not found.")
+
 class ModelFile(JsonFileResource):
     def __init__(self, data: dict = None, file_path: str = None, pack: Pack = None) -> None:
         super().__init__(data = data, file_path = file_path, pack = pack)
         self.__models = []
 
+    @property
+    def format_version(self):
+        return self.get_jsonpath("format_version")
+    
+    @format_version.setter
+    def format_version(self, format_version):
+        return self.set_jsonpath("format_version", format_version)
+
     @cached_property
     def models(self) -> list[Model]:
         for path, data in self.get_data_at("minecraft:geometry"):
@@ -2245,10 +2294,12 @@ class RenderController(JsonSubResource):
     def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
         super().__init__(data=data, parent=parent, json_path=json_path)
 
+
 class Material(JsonSubResource):
     def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
         super().__init__(data=data, parent=parent, json_path=json_path)
 
+
 class Cube(JsonSubResource):
     def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
         super().__init__(data=data, parent=parent, json_path=json_path)
@@ -2259,6 +2310,44 @@ class AnimationControllerStateRP(JsonSubResource):
         super().__init__(data=data, parent=parent, json_path=json_path)
 
 
+class TextureFileDouble(JsonSubResource):
+    """
+    A special sub-resource, which represents a texture within a texture
+    definition file, such as 'item_texture.json'
+
+    This is a special case, as it has some additional logic for obscuring the
+    texture path. This is because textures are stored nested under 
+    'textures' which is simply inconvenient to work with.
+    """
+
+    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
+        super().__init__(data=data, parent=parent, json_path=json_path)
+ 
+        self.textures = data.get("textures", [])
+        if not isinstance(self.textures, list):
+            self.textures = [self.textures]
+
+        self.textures = convert_to_notify_structure(self.textures, self)
+
+    @property
+    def shortname(self):
+        return self.id
+
+    @shortname.setter
+    def shortname(self, shortname):
+        self.id = shortname
+    
+    @property
+    def data(self):
+        """
+        Custom data getter allows us to re-create the json structure based
+        on the saved textures.
+        """
+        return {
+            "textures": self.textures
+        }
+
+
 class LootTablePool(JsonSubResource):
     def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
         super().__init__(data=data, parent=parent, json_path=json_path)
@@ -2284,13 +2373,13 @@ class AnimationControllerRP(JsonSubResource):
     def initial_state(self, initial_state):
         return self.set_jsonpath("initial_state", initial_state)
 
-    
     def get_state(self, id:str) -> AnimationControllerStateRP:
         for child in self.states:
             if smart_compare(child.id, id):
                 return child
         raise AssetNotFoundError(id)
-    
+
+
 class Model(JsonSubResource):
     def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
         super().__init__(data=data, parent=parent, json_path=json_path)
@@ -2303,7 +2392,15 @@ class Model(JsonSubResource):
             self.__bones.append(Bone(parent = self, json_path = path, data = data))
         return self.__bones
     
-    
+    def get_bone(self, name:str) -> Bone:
+        """
+        Gets a Bone via its name field.
+        """
+        for bone in self.bones:
+            if smart_compare(bone.name, name):
+                return bone
+        raise AssetNotFoundError(f"Bone with name {name} not found.")
+
     @property
     def identifier(self):
         return self.get_jsonpath("description/identifier")
@@ -2394,8 +2491,15 @@ class Bone(JsonSubResource):
         super().__init__(data=data, parent=parent, json_path=json_path)
         
         self.__cubes = []
-        
     
+    @property
+    def name(self):
+        return self.get_jsonpath("name")
+    
+    @name.setter
+    def identifier(self, name):
+        return self.set_jsonpath("name", name)
+
     @cached_property
     def cubes(self) -> list[Cube]:
         for path, data in self.get_data_at("cubes"):
