@@ -2,17 +2,28 @@ import unittest
 import sys
 import functools
 import shutil
+from typing import Union, Tuple
 
 sys.path.insert(0, '../reticulator')
 from reticulator import *
 
-def get_packs():
-    project = Project('./content/bp/', './content/rp/')
-    project.resource_pack.output_path = './out/rp/'
-    project.behavior_pack.output_path = './out/bp/'
-    return project.behavior_pack, project.resource_pack
+def get_packs() -> Tuple[BehaviorPack, ResourcePack]:
+    return Project('./content/bp/', './content/rp/').get_packs()
 
-def get_saved_packs():
+def save_and_return_packs(rp: ResourcePack = None, bp: BehaviorPack = None, force: bool = False):
+    # Prepare folder location
+    prepare_output_directory()
+
+    # Save the old packs
+    if rp is not None:
+        rp.output_path = './out/rp/'
+        rp.save(force=force)
+    
+    if bp is not None:
+        bp.output_path = './out/bp/'
+        bp.save(force=force)
+
+    # Return the saved packs packs
     project = Project('./out/bp/', './out/rp/')
     return project.behavior_pack, project.resource_pack
 
@@ -24,6 +35,51 @@ def prepare_output_directory():
 
     os.mkdir('./out')
 
+
+class TestFogs(unittest.TestCase):
+    def setUp(self) -> None:
+        self.bp, self.rp = get_packs()
+
+    def test_fog_files(self):
+        self.assertEqual(len(self.rp.fogs), 3)
+        fog_file = self.rp.get_fog('minecraft:fog_mushroom_island_shore')
+        self.assertEqual(len(fog_file.distance_components), 2)
+        self.assertEqual(len(fog_file.volumetric_density_components), 1)
+        self.assertEqual(len(fog_file.volumetric_media_coefficients), 2)
+
+        component = fog_file.get_distance_component('air')
+        self.assertEqual(component.data['fog_end'], 60)
+
+
+class TestStandaloneTextureFiles(unittest.TestCase):
+    def setUp(self) -> None:
+        self.bp, self.rp = get_packs()
+    
+    def test_terrain_texture_file(self):
+        terrain_texture = self.rp.terrain_texture_file
+        texture_definition = terrain_texture.texture_definitions[0]
+
+        self.assertEqual(len(terrain_texture.texture_definitions), 5)
+        self.assertEqual(texture_definition.textures[0], 'textures/blocks/planks_acacia')
+
+        texture_definition = terrain_texture.get_texture_definition('anvil_base')
+
+        self.assertEqual(texture_definition.textures[0], 'textures/blocks/anvil_base')
+        self.assertEqual(len(texture_definition.textures), 4)
+
+
+    def test_item_texture_file(self):
+        item_texture_file = self.rp.item_texture_file
+        texture_definition = item_texture_file.texture_definitions[0]
+
+        self.assertEqual(len(item_texture_file.texture_definitions), 5)
+        self.assertEqual(texture_definition.textures[0], 'textures/items/apple')
+    
+        texture_definition = item_texture_file.get_texture_definition('axe')
+
+        self.assertEqual(texture_definition.textures[0], 'textures/items/wood_axe')
+        self.assertEqual(len(texture_definition.textures), 6)
+
 class TestRenderControllers(unittest.TestCase):
     def setUp(self) -> None:
         self.bp, self.rp = get_packs()
@@ -219,6 +275,8 @@ class TestDirty(unittest.TestCase):
         self.function = self.bp.get_function('functions/kill_all_safe.mcfunction')
         self.command = self.function.commands[0]
         self.component = self.entity.get_component('minecraft:type_family')
+        self.item_texture_file = self.rp.item_texture_file
+        self.texture_definition = self.item_texture_file.get_texture_definition('axe')
 
     def assert_dirties(attribute):
         """
@@ -274,6 +332,20 @@ class TestDirty(unittest.TestCase):
     def test_subresource_id(self):
         self.component.id = 'new_component_name'
 
+    @assert_dirties('item_texture_file')
+    @assert_dirties('texture_definition')
+    def test_texture_definition(self):
+        self.texture_definition.shortname = 'new_shortname'
+
+    @assert_dirties('item_texture_file')
+    @assert_dirties('texture_definition')
+    def test_texture_definition_textures(self):
+        self.texture_definition.textures.append('new_texture')
+
+    @assert_dirties('item_texture_file')
+    def test_add(self):
+        self.item_texture_file.add_texture_definition('new_definition', [])
+
 class TestParticle(unittest.TestCase):
     def setUp(self) -> None:
         self.bp, self.rp = get_packs()
@@ -355,12 +427,9 @@ class TestAnimationTriple(unittest.TestCase):
         animation.identifier = 'new_identifier'
 
         # Save the resource
-        prepare_output_directory()
-        self.rp.save()
+        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)
 
-        # Test the saved resource
-        out_bp, out_rp = get_saved_packs()
-        animation = out_rp.get_entity('minecraft:dolphin').animations[0]
+        animation = saved_rp.get_entity('minecraft:dolphin').animations[0]
 
         self.assertEqual(animation.shortname, 'new_shortname')
         self.assertEqual(animation.identifier, 'new_identifier')
@@ -393,6 +462,103 @@ class TestEntityFileRP(unittest.TestCase):
         # Check filepath
         self.assertEqual(self.entity.file_path, 'entity\\dolphin.entity.json')
 
+class TestDeletion(unittest.TestCase):
+    def setUp(self) -> None:
+        self.project = Project('./content/bp/', './content/rp/')
+        self.bp = self.project.behavior_pack
+        self.rp = self.project.resource_pack
+
+    def test_file_deletion(self):
+        """
+        Tests that we can delete a file.
+        """
+
+        # Delete entity
+        entity = self.bp.get_entity('minecraft:dolphin')
+        entity.delete()
+
+        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)
+
+        # Check that the file is gone
+        with self.assertRaises(AssetNotFoundError):
+            saved_bp.get_entity('minecraft:dolphin')
+
+    def test_subresource_deletion(self):
+        # Delete entity
+        entity = self.bp.get_entity('minecraft:dolphin')
+        component = entity.get_component('minecraft:type_family')
+        component.delete()
+
+        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)
+        entity = saved_bp.get_entity('minecraft:dolphin')
+
+        with self.assertRaises(AssetNotFoundError):
+            component = entity.get_component('minecraft:type_family')
+    
+    @unittest.skip("Not implemented yet")
+    def test_list_subresource_deletion(self):
+        model = self.rp.get_model('geometry.dolphin')
+        self.assertEqual(len(model.bones), 9)
+        bone = model.get_bone('bristle2')
+
+        print("Deleting: " + bone.json_path)
+        bone.delete()
+
+        # This fails, because the list can no longer be accessed
+        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)
+
+        # bone = model.get_bone('bristle2')
+
+
+class TestModels(unittest.TestCase):
+    def setUp(self) -> None:
+        self.bp, self.rp = Project('./content/bp/', './content/rp/').get_packs()
+
+    def test_model_file(self):
+        self.assertEqual(len(self.rp.models), 1)
+        self.assertTrue(len(self.rp.model_files), 1)
+        
+        model_file = self.rp.model_files[0]
+        model = self.rp.get_model('geometry.dolphin')
+        self.assertEqual(model_file.models[0].id, model.id)
+
+        self.assertEqual(model.file.file_name, 'dolphin.geo.json')
+
+    def test_bones(self):
+        model = self.rp.get_model('geometry.dolphin')
+        self.assertEqual(len(model.bones), 9)
+
+class TestLanguageFiles(unittest.TestCase):
+    def setUp(self) -> None:
+        self.bp, self.rp = get_packs()
+
+    def test_language_file(self):
+        self.assertEqual(len(self.rp.language_files), 1)
+
+        language_file = self.rp.get_language_file('texts/es_ES.lang')
+
+        with self.assertRaises(AssetNotFoundError):
+            self.rp.get_language_file('texts/dne.lang')
+
+        self.assertEqual(len(language_file.translations), 3)
+        translation = language_file.get_translation('accessibility.text.period')
+
+        self.assertEqual(translation.key, 'accessibility.text.period')
+        self.assertEqual(translation.value, 'Punto')
+        self.assertEqual(translation.comment, '')
+
+    def test_adding_translation(self):
+        language_file = self.rp.get_language_file('texts/es_ES.lang')
+        language_file.add_translation(Translation('new_key', 'new_value'))
+        language_file.add_translation(Translation('new_key2', 'new_value2', comment="Test"))
+
+        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)
+
+        language_file = saved_rp.get_language_file('texts/es_ES.lang')
+        self.assertEqual(len(language_file.translations), 5)
+        translation = language_file.get_translation('new_key2')
+        self.assertEqual(translation.comment, 'Test')
+
 class TestJsonPathAccess(unittest.TestCase):
     """
     Test various jsonpath access methods.
